import logging
import weakref
from abc import abstractmethod
from typing import TYPE_CHECKING

from eventspype.sub.multisubscriber import MultiSubscriber

if TYPE_CHECKING:
    from agentspype.agent.agent import Agent


class AgentListening(MultiSubscriber):
    def __init__(self, agent: "Agent") -> None:
        super().__init__()
        self._agent = weakref.ref(agent)

    # === Properties ===

    @property
    def agent(self) -> "Agent":
        agent = self._agent()
        if agent is None:
            raise RuntimeError("Agent has been deactivated")
        return agent

    def logger(self) -> logging.Logger:
        return self.agent.logger()

    # === Subscriptions ===

    @abstractmethod
    def subscribe(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def unsubscribe(self) -> None:
        raise NotImplementedError



from pydantic import BaseModel, ConfigDict


class AgentConfiguration(BaseModel):
    """Base configuration class for agents."""

    model_config = ConfigDict(frozen=False, arbitrary_types_allowed=False)



import weakref
from abc import abstractmethod
from typing import TYPE_CHECKING, Any, ClassVar, TypeVar

from statemachine import State, StateMachine
from statemachine.factory import StateMachineMetaclass
from statemachine.transition_list import TransitionList

from agentspype.agent.publishing import StateAgentPublishing

if TYPE_CHECKING:
    from agentspype.agent.agent import Agent


T = TypeVar("T", bound="AgentStateMachine")


class AgentStateMachineMeta(StateMachineMetaclass):
    """Metaclass for AgentStateMachine that ensures proper state machine inheritance."""

    def __new__(
        mcs, name: str, bases: tuple[type, ...], namespace: dict[str, Any]
    ) -> type:
        # Don't modify the base class
        if name == "AgentStateMachine":
            return super().__new__(mcs, name, bases, namespace)

        # Create state definitions if not already defined
        if "starting" not in namespace:
            namespace["starting"] = State("Starting", initial=True)
        if "idle" not in namespace:
            namespace["idle"] = State("Idle")
        if "end" not in namespace:
            namespace["end"] = State("End", final=True)

        # Create transition definitions if not already defined
        if "start" not in namespace:
            namespace["start"] = namespace["starting"].to(namespace["idle"])
        if "stop" not in namespace:
            namespace["stop"] = namespace["starting"].to(namespace["end"]) | namespace[
                "idle"
            ].to(namespace["end"])

        return super().__new__(mcs, name, bases, namespace)


class AgentStateMachine(StateMachine, metaclass=AgentStateMachineMeta):
    """Base class for all agent state machines.

    This class provides the basic state machine structure that all agents should inherit from.
    Child classes can override states and transitions by defining their own, or inherit the defaults.
    """

    # === States ===

    starting: ClassVar[State]
    idle: ClassVar[State]
    end: ClassVar[State]

    # === Transitions ===

    start: ClassVar[TransitionList]
    stop: ClassVar[TransitionList]

    def __init__(self, agent: "Agent") -> None:
        super().__init__()
        self._agent = weakref.ref(agent)
        self._should_stop = False

    # === State actions ===

    def on_enter_end(self) -> None:
        self.agent.teardown()

    # === Transitions Actions ===

    def before_transition(
        self, event: str, state: str, source: str, target: str
    ) -> None:
        if source == target:
            return

        self.agent.logger().debug(
            f"[{self.agent.__class__.__name__}:StateMachine] ({event}) {source} -> {target}"
        )

    @abstractmethod
    def after_transition(self, event: str, state: State) -> None:
        """Handle behavior after a transition.

        This method should be implemented in child classes to handle post-transition behavior.
        Example implementation:
            self.agent.publishing.publish_transition(event, state)
        """
        raise NotImplementedError

    def on_start(self) -> None:
        self.agent.listening.subscribe()

    def on_stop(self) -> None:
        self._should_stop = True

    # === Conditions ===

    def should_stop(self) -> bool:
        return self._should_stop

    # === Properties ===

    @property
    def agent(self) -> "Agent":
        agent = self._agent()
        if agent is None:
            raise RuntimeError("Agent has been deactivated")
        return agent

    # === Functions ===

    def safe_start(self) -> None:
        if not self.current_state.initial:
            return
        self.start(f=True)

    def safe_stop(self) -> None:
        """Safely stop the state machine if not already in final state."""
        if self.current_state.final:
            return
        self.stop(f=True)


# Example of how to create a concrete state machine:
class BasicAgentStateMachine(AgentStateMachine):
    """A basic implementation of an agent state machine.

    This class inherits all the default states and transitions from AgentStateMachine.
    It only needs to implement the required after_transition method.
    """

    def after_transition(self, event: str, state: State) -> None:
        if isinstance(self.agent.publishing, StateAgentPublishing):
            self.agent.publishing.publish_transition(event, state)



import weakref
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any

from eventspype.pub.multipublisher import MultiPublisher
from eventspype.pub.publication import EventPublication
from statemachine import State

if TYPE_CHECKING:
    from agentspype.agent.agent import Agent


class AgentPublishing(MultiPublisher):
    def __init__(self, agent: "Agent") -> None:
        super().__init__()
        self._agent = weakref.ref(agent)

    # === Properties ===

    @property
    def agent(self) -> "Agent":
        agent = self._agent()
        if agent is None:
            raise RuntimeError("Agent has been deactivated")
        return agent


class StateAgentPublishing(AgentPublishing):
    # === Definitions ===

    class Events(Enum):
        StateMachineTransition = "sm_transition_event"

    @dataclass
    class StateMachineEvent:
        event: Any
        new_state: State

    # === Publications ===

    sm_transition_event = EventPublication(
        event_tag=Events.StateMachineTransition, event_class=StateMachineEvent
    )

    # === Events ===

    def publish_transition(self, event: Any, state: State) -> None:
        """
        Publishes a state machine transition event.

        It is recommended to call this method from the state machine.
        Example:
            def after_transition(self, event, state):
                self.agent.publishing.publish_transition(event, state)
        """
        self.publish(
            self.sm_transition_event,
            self.sm_transition_event.event_class(event, state),
        )



from pydantic import BaseModel, ConfigDict

from agentspype.agent.configuration import AgentConfiguration
from agentspype.agent.listening import AgentListening
from agentspype.agent.publishing import AgentPublishing
from agentspype.agent.state_machine import AgentStateMachine
from agentspype.agent.status import AgentStatus


class AgentDefinition(BaseModel):
    """Definition of an agent's components and configuration."""

    model_config = ConfigDict(frozen=True)

    state_machine_class: type[AgentStateMachine]
    events_listening_class: type[AgentListening]
    events_publishing_class: type[AgentPublishing]
    configuration_class: type[AgentConfiguration]
    status_class: type[AgentStatus]



import logging
from typing import TYPE_CHECKING, Any

from agentspype.agency import Agency
from agentspype.agent.configuration import AgentConfiguration
from agentspype.agent.definition import AgentDefinition

if TYPE_CHECKING:
    from agentspype.agent.listening import AgentListening
    from agentspype.agent.publishing import AgentPublishing
    from agentspype.agent.state_machine import AgentStateMachine
    from agentspype.agent.status import AgentStatus


class Agent:
    _logger: logging.Logger | None = None

    # === Definition ===

    definition: AgentDefinition

    # === Initialization ===

    def __init__(self, configuration: AgentConfiguration | dict[str, Any]):
        self._configuration = (
            configuration
            if isinstance(configuration, AgentConfiguration)
            else self.definition.configuration_class(**configuration)
        )

        self._events_publishing = self.definition.events_publishing_class(self)
        self._events_listening = self.definition.events_listening_class(self)
        self._state_machine = self.definition.state_machine_class(self)
        self._status = self.definition.status_class()

        Agency.register_agent(self)

        self.initialize()

    def initialize(self) -> None:
        pass

    def teardown(self) -> None:
        self.listening.unsubscribe()
        Agency.deregister_agent(self)

    def clone(self) -> "Agent":
        return self.__class__(self.configuration)

    # === Class Methods ===

    @classmethod
    def logger(cls) -> logging.Logger:
        if cls._logger is None:
            cls._logger = logging.getLogger(cls.__name__)
        return cls._logger

    # === Properties ===

    @property
    def configuration(self) -> AgentConfiguration:
        return self._configuration

    @property
    def machine(self) -> "AgentStateMachine":
        return self._state_machine

    @property
    def listening(self) -> "AgentListening":
        return self._events_listening

    @property
    def publishing(self) -> "AgentPublishing":
        return self._events_publishing

    @property
    def status(self) -> "AgentStatus":
        return self._status



import asyncio
from typing import TYPE_CHECKING

from bidict import bidict

if TYPE_CHECKING:
    from agentspype.agent.agent import Agent
    from agentspype.agent.configuration import AgentConfiguration


class Agency:
    initialized_agents: list["Agent"] = []
    _deactivating_agents: list["Agent"] = []
    _agent_to_configuration: bidict[type["Agent"], type["AgentConfiguration"]] = (
        bidict()
    )

    @classmethod
    def register_agent(cls, agent: "Agent") -> None:
        """Register an initialized agent."""
        if agent not in cls.initialized_agents:
            agent.logger().info(f"[Agency] Registered: {agent.__class__.__name__}")
            cls.initialized_agents.append(agent)

    @classmethod
    def deregister_agent(cls, agent: "Agent") -> None:
        """Deregister an initialized agent."""
        if agent in cls.initialized_agents:
            agent.logger().info(f"[Agency] Deregistered: {agent.__class__.__name__}")
            cls.initialized_agents.remove(agent)
            cls._deactivating_agents.append(agent)

            async def cleanup() -> None:
                await asyncio.sleep(0)
                cls._deactivating_agents.remove(agent)

            asyncio.ensure_future(cleanup())

    @classmethod
    def get_active_agents(cls) -> list["Agent"]:
        """Get a list of all active agents."""
        return list(cls.initialized_agents)

    @classmethod
    def register_agent_class(cls, agent_class: type["Agent"]) -> None:
        """Register an agent class from its configuration."""
        if agent_class in cls._agent_to_configuration:
            return

        cls._agent_to_configuration[agent_class] = (
            agent_class.definition.configuration_class
        )

    @classmethod
    def deregister_agent_class(cls, agent_class: type["Agent"]) -> None:
        """Deregister an agent class."""
        if agent_class not in cls._agent_to_configuration:
            return

        del cls._agent_to_configuration[agent_class]

    @classmethod
    def get_agent_from_configuration(
        cls, configuration: "AgentConfiguration"
    ) -> "Agent":
        """Get an agent from its configuration."""
        if type(configuration) not in cls._agent_to_configuration.inverse:
            raise ValueError(
                f"No agent class registered for configuration {type(configuration)}"
            )

        agent_class = cls._agent_to_configuration.inverse[type(configuration)]
        return agent_class(configuration)
